import os
import json
from datetime import datetime, timezone
import re
# import traceback
# import pathlib
from splunk.persistconn.application import PersistentServerConnectionApplication
from splunk.clilib import cli_common as cli
import splunk.mining.dcutils as dcu

class Investigation(PersistentServerConnectionApplication):
    """
    TODO
        Array with dicts: index, search and workflowID
    Example in splunk with the script investigation.js:
    <input type="checkbox" token="hayabusaField" searchWhenChanged="false">
      <label>Hayabusa</label>
      <delimiter> </delimiter>
      <choice value=""> </choice>
      <change>
        <condition value="">
          <set token="hayabusaSource">sourcetype=ld/json</set>
          <set token="source_1">{
          "index": $indexCombobox|s$,
          "search": $hayabusaSource|s$,
          "workflowID": "b5deffdd-f4b9-4d1a-8ea0-9b3410d915e7_PoC21:372fc5be-2314-4b7a-8fa2-d5dc70708a4b_Investigationhighlevel"
          }</set>
        </condition>
      </change>
    </input>
    """
    _sources = []

    """
    TODO
    """
    _inferences = []
    _indexes = []

    """
    TODO
        Dict with keys:
            action: create or delete
            level: low or high
            sources: [],
            inferences: ""
            investigationHighLevelId: "" when create an investigation Low Level
    """
    _params = {}

    """
    TODO
       logs are in python.log
       Get-Content "C:\Program Files\Splunk\var\log\splunk\python.log" -Wait -Tail 30
       tail -f /opt/cmem-orchestration/data/splunk/var/log/splunk/python.log
    """
    _log = dcu.getLogger()

    """
    TODO
    """
    _storateType=""
    """
    TODO
    """
    _cfgStorage=""
    """
    TODO
    """
    _splunkProtocol=""
    """
    TODO
    """
    _splunkHost=""
    """
    TODO
    """
    _splunkPort=""
    """
    TODO
    """
    _splunkToken=""
    """
    TODO
    """
    _convertJsonlToJson=""
    """
    TODO
    """
    _diMountpoint=""

    def __init__(self, _command_line, _command_arg):
        super(PersistentServerConnectionApplication, self).__init__()
        # read settings
        self._log.debug("init read settings") 
        cfgStorageSettings = cli.getConfStanza('settings','workflow:storage')
        self._storateType = cfgStorageSettings.get('type')
        self._cfgStorage = cli.getConfStanza('settings',self._storateType)

        cfgCmemSettings = cli.getConfStanza('settings','workflow:cmem')
        self._diMountpoint =  cfgCmemSettings.get('diMountpoint')
        
        cfgSplunkSettings = cli.getConfStanza('settings','workflow:splunk')
        self._splunkProtocol = cfgSplunkSettings.get('protocol')
        self._splunkHost = cfgSplunkSettings.get('host')
        self._splunkPort = cfgSplunkSettings.get('port')
        self._splunkToken = cfgSplunkSettings.get('token')

        cfgSplunkSettings = cli.getConfStanza('settings','workflow:tools')
        self._convertJsonlToJson = cfgSplunkSettings.get('convertJsonlToJson')
        self._log.debug("init finished") 
        
    # Handle a syncronous from splunkd.
    def handle(self, in_string):
        """
        Called for a simple synchronous request.
        @param in_string: request data passed in
        @rtype: string or dict
        @return: String to return in response.  If a dict was passed in,
                 it will automatically be JSON encoded before being returned.
        """

        #read parameters
        input = json.loads(in_string.decode('utf-8'))

        self._sources = []
        self._inferences = [] 
        self._log.debug("read parameters") 
        for param in input["query"]:
            self._log.debug(param[0] + "=" + param[1]) 
            if param[0] == "sources":
                try:
                    self._sources.append(json.loads(param[1]))
                except ValueError:
                    message = "Check parameter: sources (JSON error):" + param[1]
                    self._log.debug(message) 
                    payload = {
                        "message": message
                    }
                    return {'payload': payload, 'status': 400}
            elif param[0] == "inferences":
                try:
                    if len(param[1]) > 0:
                        self._inferences= json.loads(param[1])
                except ValueError:
                    message = "Check parameter: inferences (JSON error):" + param[1]
                    self._log.debug(message) 
                    payload = {
                        "message": message
                    }
                    return {'payload': payload, 'status': 400}
            elif param[0] == "indexes":
                try:
                    self._indexes= re.findall( r"index=\"([^\"]+)\"", param[1])
                except ValueError:
                    message = "Check parameter: indexes (regex error): "+ param[1]
                    self._log.debug(message) 
                    payload = {
                        "message": message
                    }
                    return {'payload': payload, 'status': 400}
            else:
                self._params[param[0]] = param[1]
        self._log.debug("read parameters finished") 

        if "action" in self._params:
            if self._params["action"] == "create":
                self._log.debug("create") 
                return self._create()
            elif self._params["action"] == "delete":
                self._log.debug("delete") 
                return self._delete()
            else: 
                message = "Action unknown:" +self._params["action"]
                self._log.debug(message) 
                payload = {
                    "message": message
                }
                return {'payload': payload, 'status': 400}

    def handleStream(self, handle, in_string):
        """
        For future use
        """
        raise NotImplementedError(
            "PersistentServerConnectionApplication.handleStream")

    def done(self):
        """
        Virtual method which can be optionally overridden to receive a
        callback after the request completes.
        """
        pass

    def _writeFile(self,filename,content):
        # replacement strings
        WINDOWS_LINE_ENDING = b'\r\n'
        UNIX_LINE_ENDING = b'\n'
        # # Windows ➡ Unix
        scriptBashBytes = str.encode(content).replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

        # self._log.info("INVESTIGATION STORAGE: "+self._storateType)
        if self._storateType == "local":
            file = os.path.join(self._cfgStorage.get('pathInvestigationInSplunk'), filename)

            with open(file, 'wb') as f:
                f.write(scriptBashBytes)

    def _buildTTL(self,id,level,date,description,isFinished,graphes=[],indexes=[],investigationHighLevelId=""):
        self._log.debug("_buildTTL started") 
        type = "cti:Investigation"+level.capitalize()+"Level";
        label = "Investigation {level}-level".format(level=level)
        dateiso= date.isoformat().replace("+00:00", "Z")
        status = "cti:InProgress"
        ttlgraphesValue =""
        if len(graphes) > 0:
            ttlgraphesValue = "owl:imports  <" + ">, <".join(graphes) + ">  ;"
        ttlindexesValue =""
        if len(indexes) > 0:
            ttlindexesValue = "cti:indexSelectedInSplunk  \\\"" + "\\\", \\\"".join(indexes) + "\\\"  ;"
        if isFinished:
            status = "cti:Ready"

        investigationHighLevelIdTTL = ""
        if len(investigationHighLevelId) > 0:
            investigationHighLevelIdTTL = "cti:investigationHighLevelId  "+investigationHighLevelId+" ;"

        ttl = '''
prefix owl:  <http://www.w3.org/2002/07/owl#>
prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
prefix xsd:  <http://www.w3.org/2001/XMLSchema#>
prefix cti:  <https://osint.eccenca.dev/cti#>

<${{investigation_graph}}>
        rdf:type     <http://rdfs.org/ns/void#Dataset> , cti:Investigation, {type} ;
        rdfs:label   \\"\\"\\"{label}\\"\\"\\"  ;
        rdfs:description   \\"\\"\\"{description}\\"\\"\\"  ;
        cti:dateTime \\"{datetime}\\"^^xsd:dateTime ;
        cti:id {id} ;
        {investigationHighLevelIdTTL}
        {ttlgraphesValue}
        {ttlindexesValue}
        cti:status   {status} . 
        '''.format(
            type = type,
            label = label,
            description = description.replace("\"", "\\\""),
            datetime= dateiso,
            status = status,
            ttlgraphesValue=ttlgraphesValue,
            id=id,
            ttlindexesValue = ttlindexesValue,
            investigationHighLevelIdTTL=investigationHighLevelIdTTL
        )
        self._log.debug("_buildTTL finished:" + ttl) 
        return ttl

    def _create(self):
        self._log.debug("_create started") 

        if len(self._sources) == 0 :
            payload = {
                "message": "No sourcetype selected"
            }
            return {'payload': payload, 'status': 400}

        investigationId = datetime.now().strftime("%Y%m%d%H%M%S%f")
        investigationFilename = "investigation_" + investigationId + ".sh"        
           
        scriptBash = '''\
#!/bin/bash -x
# FILE : {filename}
# INVESTIGATION {id} : {level} {action}

id={id}
di_mountpoint={diMountpoint}
protocol={protocol}
host={host}
port={port}
token={token}

output_dir_data={outputDirData}/${{id}}

mkdir -p ${{output_dir_data}}
cd ${{output_dir_data}}
'''.format(
            outputDirData=self._cfgStorage.get('pathData'),
            id = investigationId,
            filename = investigationFilename,
            level = self._params["level"],
            action = self._params["action"],
            protocol= self._splunkProtocol,
            host=self._splunkHost,
            port=self._splunkPort,
            token=self._splunkToken,
            diMountpoint=self._diMountpoint
        )
        
        date = datetime.now(timezone.utc)
        scriptBash = scriptBash + '''
# create the graph to manage other graphs
investigation_graph=http://example.com/investigation/${{id}}
echo "
{ttl}
" > investigation_${{id}}.ttl

echo "${{investigation_graph}}" > investigation_${{id}}.ttl.graph

cmemc graph import ${{output_dir_data}}
        '''.format(
            ttl=self._buildTTL(investigationId, self._params["level"], date, "", False, [], self._indexes, self._params["investigationHighLevelId"])
        )

        limit = " " 
        graphes = []
        desc ="";
        for key, source in enumerate(self._sources):
            
            # self._log.info("INVESTIGATION..OK..test1") 

            spl = "search " + source["index"] + " " + source["search"] + limit
            desc = desc + "Source {key}: {spl}\n".format(key=key,spl=spl)
            #TODO implement or find a real sanitizer for SPL language
            if "rm " in spl or "cd " in spl or "'" in spl or "\"\"\"" in spl :
                
                # self._log.info("INVESTIGATION..OK..test2") 
                payload = {
                    "message": "SPL sanitizer: problems found in the search: "+spl
                }
                return {'payload': payload, 'status': 400}

            self._log.info("INVESTIGATION..OK..debug9") 
            graph = "http://example.com/investigation/{id}/{key}".format(
                    id = investigationId,
                    key = key
                )
            graphes.append(graph)
            
            scriptBash = scriptBash + '''
# search {search}
search='{search}'

output_file_splunk_json=source_{key}.json
output_file_cmemc=graph_{key}.ttl

curl -k  ${{protocol}}://${{host}}:${{port}}/servicesNS/-/search/search/jobs/export\\
     -H "Authorization: Splunk ${{token}}"\\
     --data-urlencode search="${{search}}" -d output_mode=raw  \\
        {convertJsonlToJson} > ${{output_file_splunk_json}}

#PLAN A: ERROR 502
# cmemc workflow io {workflowID} -i ${{output_file_splunk_json}} -o ${{output_file_cmemc}}.ttl

#PLAN B:
# copy the (big) file locally instead of http transfer
cp "${{output_file_splunk_json}}" ${{di_mountpoint}}/{projectID}/resources/source_tmp.json

# query very slow with "DROP" in GraphDB
# drop tmp graph (DROP GRAPH <{{graph_iri}}> ;)
cmemc query execute https://ns.eccenca.com/data/queries/a34ff902-d468-459a-b8a7-c0ccce8b67a4 --parameter graph_iri "https://osint.eccenca.dev/tmp/"

cmemc workflow execute --wait {workflowID}

# query very slow with "MOVE TO" in GraphDB
cmemc query execute https://ns.eccenca.com/data/queries/f8c58b93-4719-4f79-a848-c28a1523ba95 --parameter graph {graph}
cmemc query execute https://ns.eccenca.com/data/queries/a34ff902-d468-459a-b8a7-c0ccce8b67a4 --parameter graph_iri "https://osint.eccenca.dev/tmp/"

#PLAN A:
# Maybe useful if we want to save also the ttl. echo "{graph}" > ${{output_file_cmemc}}.ttl.graph

'''.format(
        search = spl,
        projectID = source["workflowID"].split(':')[0],
        workflowID = source["workflowID"],
        id = investigationId,
        key = key,
        graph = graph,
        convertJsonlToJson = self._convertJsonlToJson
    )
        scriptBash = scriptBash + '''

# create the graph to manage other graphs
investigation_graph=http://example.com/investigation/${{id}}
echo "
{ttl}
" > investigation_${{id}}.ttl

echo "${{investigation_graph}}" > investigation_${{id}}.ttl.graph

cmemc graph import --replace ${{output_dir_data}}

'''.format(
        ttl=self._buildTTL(investigationId, self._params["level"], date, desc, True, graphes, self._indexes, self._params["investigationHighLevelId"])
    )

        # Calculate inferences
        for key, inference in enumerate(self._inferences):
            if(inference["command"] == "query"):
                scriptBash = scriptBash + '''
cmemc query execute -p investigationID {investigationID} -p inferenceID {inferenceID} {iri}
'''.format(investigationID=investigationId,inferenceID=inference["inferenceID"],iri=inference["queryIRI"])
            elif(inference["command"] == "workflow"):
                scriptBash = scriptBash + '''
cmemc workflow execute --wait {workflowID} 
'''.format(workflowID=inference["workflowID"])

        try:
            self._writeFile(investigationFilename,scriptBash)
        except Exception as e:
            self._log.exception(e)
            payload = {
                "action": "create",
                "message": "Error: see the logs"
            }
            return {'payload': payload, 'status': 500}

        # payload = {
        #     "file": file,
        #     "script": scriptBash
        # }        
        payload = {
            "action": "create"
        }
        self._log.debug("_create finished") 
        return {'payload': payload, 'status': 200}

    def _delete(self):
        self._log.debug("_delete started") 
        if "investigationID" not in self._params :
            payload = {
                "action": "delete",
                "message": "No investigation selected"
            }
            return {'payload': payload, 'status': 400}

        investigationFilenameDelete = "investigation_" + self._params["investigationID"] + ".del"  
        try:
            self._writeFile(investigationFilenameDelete, self._params["investigationID"])
        except Exception as e:
            self._log.exception(e)
            payload = {
                "action": "delete",
                "message": "Error: see the logs"
            }
            return {'payload': payload, 'status': 500}

        payload = {
            "action": "delete"
        }
        self._log.debug("_delete finished") 
        return {'payload': payload, 'status': 200}
